/*-
 * Copyright (C)2002..2018 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ident "@(#) Copyright (C)2002..2018 @BABOLO http://www.babolo.ru/"
#ident "@(#) $Id: cmd_select.c,v 1.64 2018/06/12 18:11:36 babolo Exp $"

#define BLIN_COMPAT      3
#define Bpars_COMPAT     3
#define MIFE_COMPAT      4
#define PGOBLIN_COMPAT   2
#define PGOBLIN_INTERNAL 1

#include <sys/types.h>
#include <sys/event.h>
#include <sys/time.h>
#include <sysexits.h>
#include <unistd.h>
#include <stdlib.h>
#include <limits.h>
#include <string.h>
#include <errno.h>
#include <stdio.h>
#include <err.h>
#include <babolo/BLINflag.h>
#include <babolo/parser.h>
#include <mife.h>
#include "pgoblin.h"

int
pgoblin_perform(pgoblin_exenv *exenv, pgoblin_nr *rn) {
    int ex = EX_OK, i, j, row, col, *len;
    pgoblin_conn *conn;
    pgoblin_mode *res;
    const char **val;
    pgoblin_io *ctl;
    pgoblin_mode *n;
    pgoblin_io *in;

#   define blin_internal_flags (exenv->options->flags & BLIN_MASK)
    ifBLIN_QX3( "+perform .%c%c%c.."
              , pgoblin_regn[rn[PGO_CCTL]]
              , pgoblin_regn[rn[PGO_CIN]]
              , pgoblin_regn[rn[PGO_CCON]]
              );
    conn = &(exenv->options->conn[rn[PGO_CCON]]);
    ctl = &(exenv->options->io[rn[PGO_CCTL]]);
    in = &(exenv->options->io[rn[PGO_CIN]]);
    if (ctl->text) ifBLIN_QX3(" =%s", (char*)ctl->text);
    if  (!!rn[PGO_CIN] && !!in->pq) {
        if  (ctl->text) {
        } else if (!ctl->pq) {
            ifBLIN_QX0("perform: No query #1");
            ERROUT(EX_NOINPUT, EINVAL);
        } else {
            /*
             *   IO_CTL Pq,  IO_IN Pq,       
             * ,    IO_CTL Pq Pt  q     IO_IN Pq.
             */
            if  ((ex = pgoblin_style_0.table(ctl, exenv->options->style, ctl, 0, -2)) < 0) {
                ifBLIN_QW0("perform: No STYLE 0 #1");
                goto out;
        }   }
        if  (  !(col = pgoblin_db_resinfo(PGOBLIN_Nfields, in->pq))
            || !(row = pgoblin_db_resinfo(PGOBLIN_Ntuples, in->pq))
            ) goto out;
        ifBLIN_QX4("%d rows x %d cols", row, col);
        if  (!(val = malloc(sizeof(char *) * col))) {
            ifBLIN_QX0("perform: No mem #1");
            ERROUT(EX_OSERR, ENOMEM);
        }
        if  (!(len = malloc(sizeof(int) * col))) {
            ifBLIN_QX0("perform: No mem #2");
            ERROUT(EX_OSERR, ENOMEM);
        }
        if  (DBASE(conn->flags)->flags & PGOBLIN_DB_PREPARE) {
            if  ((ex = pgoblin_db_prepare(conn, &n, ctl->text))) {
                ifBLIN_QO0 {
                    char *c = pgoblin_db_subinfo(PGOBLIN_ResMess, n, 0);

                    ifBLIN_QX0(" prepare\n%s", c);
                    DBASE(conn->flags)->finish(conn->flags, PGOBLIN_FrMess, (void**)&c);
                }
                ERROUT(EX_SOFTWARE, EDOOFUS);
            }
            for (i = 0; i < row; i++) {
                for (j = 0; j < col; j++) {
                    val[j] = pgoblin_db_getvalue(in->pq, i, j);
                    len[j] = pgoblin_db_valinfo(PGOBLIN_Length, in->pq, i, j);
                }
                ex = pgoblin_db_execute(conn, NULL, n, col, val, len, PGOBLIN_CR_OK | PGOBLIN_ALTERQ);
            }
            pgoblin_db_clear(&n);
        } else {
            for (i = 0; i < row; i++) {
                for (j = 0; j < col; j++) {
                    val[j] = pgoblin_db_getvalue(in->pq, i, j);
                    len[j] = pgoblin_db_valinfo(PGOBLIN_Length, in->pq, i, j);
                }
                ex = pgoblin_db_execute(conn, NULL, ctl->text, col, val, len, PGOBLIN_CR_OK);
        }   }
        free(val);
        free(len);
    } else if (ctl->text) {
        if  ((ex = pgoblin_db_query(conn, &res, ctl->text, PGOBLIN_CR_OK))) {
            ifBLIN_QO0 {
                char *f = pgoblin_db_subinfo(PGOBLIN_ResMess, res, 0);
                ifBLIN_QX0(" query #1: %s", f);
                DBASE(conn->flags)->finish(exenv->options->flags, PGOBLIN_FrMess, (void**)&f);
            }
            ERROUT(EX_SOFTWARE, EDOOFUS);
        }
        pgoblin_db_clear(&res);
    } else if (!ctl->pq) {
        ifBLIN_QX0(" No query #2");
        ERROUT(EX_NOINPUT, EINVAL);
    } else {
        ssize_t k;

        for (k = 0; k < pgoblin_db_resinfo(PGOBLIN_Ntuples, ctl->pq); k++) {
            if  ((ex = pgoblin_style_0.table(ctl, exenv->options->style, ctl, k, -3)) < 0) {
                ifBLIN_QW0(" No STYLE 0 #2");
                goto out;
            }
            if  ((ex = pgoblin_db_query(conn, &res, ctl->text, PGOBLIN_CR_OK))) {
                ifBLIN_QO0 {
                    char *f = pgoblin_db_subinfo(PGOBLIN_ResMess, res, 0);
                    ifBLIN_QX0(" query #2: %s", f);
                    DBASE(conn->flags)->finish(exenv->options->flags, PGOBLIN_FrMess, (void**)&f);
                }
                ERROUT(EX_SOFTWARE, EDOOFUS);
            }
            pgoblin_db_clear(&res);
    }   }
out:
    ifBLIN_QX3("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}

int
pgoblin_select(pgoblin_exenv *exenv, pgoblin_nr *rn) {
    int ex = EX_OK, i, j, row, col, *len;
    pgoblin_conn *conn;
    pgoblin_io *out;
    pgoblin_io *ctl;
    pgoblin_io *in;
    const char **val;
    pgoblin_mode *n;

#   define blin_internal_flags (exenv->options->flags & BLIN_MASK)
    ifBLIN_QX3( "+select %c%c%c%c.."
              , pgoblin_regn[rn[PGO_COUT]]
              , pgoblin_regn[rn[PGO_CCTL]]
              , pgoblin_regn[rn[PGO_CIN]]
              , pgoblin_regn[rn[PGO_CCON]]
              );
    conn = &(exenv->options->conn[rn[PGO_CCON]]);
    out = &(exenv->options->io[rn[PGO_COUT]]);
    ctl = &(exenv->options->io[rn[PGO_CCTL]]);
    in = &(exenv->options->io[rn[PGO_CIN]]);
    if  (!ctl->text) {
        ifBLIN_QX0("pgoblin_select: No query");
        ex = EX_NOINPUT;
        errno = EINVAL;
    } else if (rn[PGO_CIN] && in->pq) {
        if  (  !(col = pgoblin_db_resinfo(PGOBLIN_Nfields, in->pq))
            || !(row = pgoblin_db_resinfo(PGOBLIN_Ntuples, in->pq))
            ) goto out;
        if  (!(val = malloc(sizeof(char *) * col * row))) {
            ifBLIN_QX0("pgoblin_select #1: No mem");
            ERROUT(EX_OSERR, ENOMEM);
        }
        if  (!(len = malloc(sizeof(int) * col * row))) {
            ifBLIN_QX0("pgoblin_select #2: No mem");
            ERROUT(EX_OSERR, ENOMEM);
        }
        if  (DBASE(conn->flags)->flags & PGOBLIN_DB_PREPARE) {
            if  ((ex = pgoblin_db_prepare(conn, &n, ctl->text))) {
                ifBLIN_QX0("pgoblin_select: prepare");
                ERROUT(EX_SOFTWARE, EDOOFUS);
            }
            for (i = 0; i < row; i++) {
                for (j = 0; j < col; j++) {
                    val[i * col + j] = pgoblin_db_getvalue(in->pq, i, j);
                    len[i * col + j] = pgoblin_db_valinfo(PGOBLIN_Length, in->pq, i, j);
            }   }
            MARK_IO_PQ_NGO(rn[PGO_COUT]);;
            ;;  ex = pgoblin_db_execute( conn
        /*  ;;  */                     , &out->pq
        /*  ;;  */                     , n, col * row, val, len
        /*  ;;  */                     , PGOBLIN_CR_TUPL | PGOBLIN_ALTERQ
        /*  ;;  */                     );
            ;;  out->flagc = out->pq ? conn->flags : 0;
            ;;  if  (out->pq) out->flags |= PGOBLIN_PQRESULT | PGOBLIN_FREEPQRE;
            MARK_IO_PQ_NWENT(rn[PGO_COUT]);;
            if  (ex) {
                ifBLIN_QX0("pgoblin_select: execute Q");
                ERROUT(EX_SOFTWARE, EDOOFUS);
            }
            pgoblin_db_clear(&n);
        } else {
            for (i = 0; i < row; i++) {
                for (j = 0; j < col; j++) {
                    val[i * col + j] = pgoblin_db_getvalue(in->pq, i, j);
                    len[i * col + j] = pgoblin_db_valinfo(PGOBLIN_Length, in->pq, i, j);
            }   }
            MARK_IO_PQ_NGO(rn[PGO_COUT]);;
            ;;  ex = pgoblin_db_execute( conn
        /*  ;;  */                     , &out->pq
        /*  ;;  */                     , ctl->text
        /*  ;;  */                     , col * row, val, len
        /*  ;;  */                     , PGOBLIN_CR_TUPL
        /*  ;;  */                     );
            ;;  out->flagc = out->pq ? conn->flags : 0;
            ;;  if  (out->pq) out->flags |= PGOBLIN_PQRESULT | PGOBLIN_FREEPQRE;
            MARK_IO_PQ_NWENT(rn[PGO_COUT]);;
            if  (ex) {
                ifBLIN_QX0("pgoblin_select: execute");
                ERROUT(EX_SOFTWARE, EDOOFUS);
        }   }
        free(val);
        free(len);
    } else {
        MARK_IO_PQ_NGO(rn[PGO_COUT]);;
        ;;  ex = pgoblin_db_query(conn, &out->pq, ctl->text, PGOBLIN_CR_TUPL);
        ;;  out->flagc = out->pq ? conn->flags : 0;
        ;;  if  (out->pq) out->flags |= PGOBLIN_PQRESULT | PGOBLIN_FREEPQRE;
        MARK_IO_PQ_NWENT(rn[PGO_COUT]);;
        if  (ex) {
            ifBLIN_QX0("pgoblin_select: query");
            ERROUT(EX_SOFTWARE, EDOOFUS);
    }   }
out:
    ifBLIN_QX3("- %s", out->pq ? "OK" : "NULL");
    return(ex);
#   undef blin_internal_flags
}

int
pgoblin_display(pgoblin_exenv *exenv, pgoblin_nr *rn) {
    pgoblin_styreg *sty;
    pgoblin_io *out;
    pgoblin_io *in;
    ssize_t ex;

#   define blin_internal_flags (exenv->options->flags & BLIN_MASK)
    ifBLIN_QX3( "+display %c.%c..%c"
              , pgoblin_regn[rn[PGO_COUT]]
              , pgoblin_regn[rn[PGO_CIN]]
              , pgoblin_regn[rn[PGO_CSTY]]
              );
    sty = &(exenv->options->style[rn[PGO_CSTY]]);
    out = &(exenv->options->io[rn[PGO_COUT]]);
    in = &(exenv->options->io[rn[PGO_CIN]]);
    ex = STYLS(sty->flags)->table(out, sty, in, 0, -2);
    if  (ex < 0) ifBLIN_QX0("pgoblin_display: write error");
    ifBLIN_QX3("- %"BLIN_D, ex);
    return((ex < 0) ? ex : 0);
#   undef blin_internal_flags
}

static int
sel(pgoblin_exenv *exenv, pgoblin_nr *rn, int cond) {
    ssize_t k, e = 0, ex = 0;
    pgoblin_styreg *sty;
    pgoblin_conn *conn;
    pgoblin_io *out;
    pgoblin_io *ctl;
    pgoblin_io tr;
    size_t i;
#   define aim  (INT_MAX / 2)
#   define aid                                                                                        \
    ((PGOBLIN_BUFMULT * PGOBLIN_BUFSIZE > (aim / 4)) ? (aim / 4) : (PGOBLIN_BUFMULT * PGOBLIN_BUFSIZE))

#   define blin_internal_flags (exenv->options->flags & BLIN_MASK)
    sty = &(exenv->options->style[rn[PGO_CSTY]]);
    conn = &(exenv->options->conn[rn[PGO_CCON]]);
    ctl = &(exenv->options->io[rn[PGO_CCTL]]);
    out = &(exenv->options->io[rn[PGO_COUT]]);
    if  (!ctl->text) {
        ifBLIN_QX0("pgoblin_strsel: No query");
        ERROUT(EX_DATAERR, EINVAL);
    }
    tr.flagc = conn->flags;
    if  ((ex = pgoblin_db_query(conn, &tr.pq, (char*)ctl->text, cond))) goto out;
    for (k = 0, i = 1; ; ) {
        int typn;

        if  ((ex = pgoblin_db_getstream(conn, &tr.pq, i, cond))) goto out;
        typn = pgoblin_db_resinfo(PGOBLIN_Ntuples, tr.pq);
        e = STYLS(sty->flags)->table(out, sty, &tr, k, typn ? -1 : -2);
        if  ((e < 0) || (typn <= 0)) break;
        k += typn;
        pgoblin_db_clear(&tr.pq);
        ifBLIN_QX3("e=%d=>%"BLIN_D" aim=%d i %"BLIN_D" => ", typn, e, aim, i);
        if  (e < (aim * 2)) {
            if  (i < (aid / 2)) i = i * 2; else i = aid;
        } else if (e < aim) {
            if  (i < (aid / aim * e)) i = i / e * aim; else i = aid;
        } else if ((e / 2) > aim) {
            i = i / 4;
        } else if (e > aim) {
            i = i / e * aim;
        }
        ifBLIN_QX3("%"BLIN_D, i);
    }
    pgoblin_db_clear(&tr.pq);
    if  (e < 0) ifBLIN_QX0("pgoblin_strsel: write error");
out:
/* END      ? */
    ex = pgoblin_db_endstream(conn, cond);
    ifBLIN_QX3( "- %s %"BLIN_D
              , (cond == PGOBLIN_CR_SSEL)
              ? "pgoblin_strsel"
              : (cond == PGOBLIN_CR_BSEL) ? "pgoblin_binsel" : "pgoblin_???sel"
              , (e < 0) ? e : ex
              );
    return((e < 0) ? e : ex);
#   undef blin_internal_flags
}

int
pgoblin_strsel(pgoblin_exenv *exenv, pgoblin_nr *rn) {
    int ex = EX_OK;

    ifBLIN_QX3( "+strselect %c%c.%c.%c"
              , pgoblin_regn[rn[PGO_COUT]]
              , pgoblin_regn[rn[PGO_CCTL]]
              , pgoblin_regn[rn[PGO_CCON]]
              , pgoblin_regn[rn[PGO_CSTY]]
              );
    ex = sel(exenv, rn, PGOBLIN_CR_SSEL);
    ifBLIN_QX3("- %"BLIN_D, ex);
    return(ex);
}

int
pgoblin_binsel(pgoblin_exenv *exenv, pgoblin_nr *rn) {
    int ex = EX_OK;

    ifBLIN_QX3( "+binselect %c%c.%c.%c"
              , pgoblin_regn[rn[PGO_COUT]]
              , pgoblin_regn[rn[PGO_CCTL]]
              , pgoblin_regn[rn[PGO_CCON]]
              , pgoblin_regn[rn[PGO_CSTY]]
              );
    ex = sel(exenv, rn, PGOBLIN_CR_BSEL);
    ifBLIN_QX3("- %"BLIN_D, ex);
    return(ex);
}

int
pgoblin_binpre(pgoblin_exenv *exenv, pgoblin_nr *rn) {
    pgoblin_conn *conn;
    pgoblin_mode *res;
    pgoblin_io *out;
    pgoblin_io *ctl;
    int ex = EX_OK;
    char *q;
#   define CMD1 "BEGIN"
#   define CMD2 "DECLARE pgoblin BINARY NO SCROLL CURSOR FOR %s"
#   define CMD3 "FETCH 2147483647 FROM pgoblin"
#   define CMD4 "CLOSE pgoblin"
#   define CMD5 "END"

#   define blin_internal_flags (exenv->options->flags & BLIN_MASK)
    ifBLIN_QX3( "+binpresel %c%c.%c.."
              , pgoblin_regn[rn[PGO_COUT]]
              , pgoblin_regn[rn[PGO_CCTL]]
              , pgoblin_regn[rn[PGO_CCON]]
              );
    conn = &(exenv->options->conn[rn[PGO_CCON]]);
    out = &(exenv->options->io[rn[PGO_COUT]]);
    ctl = &(exenv->options->io[rn[PGO_CCTL]]);
    if  (!ctl->text) {
        ifBLIN_QX0("pgoblin_binpre: No query");
        ERROUT(EX_DATAERR, EINVAL);
    }
    if  (!conn->intran && (ex = pgoblin_db_query(conn, &res, CMD1, PGOBLIN_CR_OK)))
        goto out;
    asprintf(&q, CMD2, (char*)ctl->text);
    if  (q == NULL) {
        ifBLIN_QX0("pgoblin_binpre: No memory");
        ERROUT(EX_OSERR, ENOMEM);
    }
    ex = pgoblin_db_query(conn, &res, q, PGOBLIN_CR_OK);
    pgoblin_db_clear(&res);
    free(q);
    if  (ex) goto out;
    MARK_IO_PQ_NGO(rn[PGO_COUT]);;
    ;;  ex = pgoblin_db_query(conn, &out->pq, CMD3, PGOBLIN_CR_TUPL);
    ;;  out->flagc = out->pq ? conn->flags : 0;
    ;;  if  (out->pq) out->flags |= PGOBLIN_PQRESULT | PGOBLIN_FREEPQRE;
    MARK_IO_PQ_NWENT(rn[PGO_COUT]);;
    ex = pgoblin_db_query(conn, &res, CMD4, PGOBLIN_CR_OK);
    pgoblin_db_clear(&res);
out:
    if  (!conn->intran) ex = pgoblin_db_query(conn, &res, CMD5, PGOBLIN_CR_OK);
    pgoblin_db_clear(&res);
    ifBLIN_QX3("- %d", 0);
    return(ex);
#   undef blin_internal_flags
}
