/*-
 * Copyright (C)2017..2018 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ident "@(#) Copyright (C)2017..2018 @BABOLO http://www.babolo.ru/"
#ident "@(#) $Id: func.c,v 1.12 2018/06/12 18:11:36 babolo Exp $"

#define BLIN_COMPAT      3
#define Bpars_COMPAT     3
#define MIFE_COMPAT      4
#define PGOBLIN_COMPAT   2
#define PGOBLIN_INTERNAL 1

#include <sys/types.h>
#include <sys/event.h>
#include <sys/time.h>
#include <unistd.h>
#include <stdio.h>
#include <errno.h>
#include <babolo/BLINflag.h>
#include <babolo/parser.h>
#include <multilar.h>
#include <mife.h>
#include "pgoblin.h"

int
/************************************************************************
 ************************************************************************
 **      cmd   rn.     **
 **    cmd  rn  ,   < 0 **
 **                                                                    **/
pgoblin_breg(pgoblin_command cmd, pgoblin_nr rn) {                    /**
 **                                                                    **
 ************************************************************************
 ************************************************************************/
    int j;

    j = -1;
    if  (pgoblin.a[cmd].flags & pgoblin.p[rn].sym) {
        /*  a[cmd].flags    [rn] */
        j = (pgoblin.p[rn].bin < 0)
            /*      */
          ? (pgoblin.a[cmd].flags >> ~pgoblin.p[rn].bin) & pgoblin.p[rn].mask
            /*      */
          : pgoblin.p[rn].bin
        ;
        if  (!(pgoblin.a[cmd].ch[j] & pgoblin.p[rn].val) && !!pgoblin.a[cmd].ch[j]) {
            /*                            *
             *     ,  -     . *
             *    0 j ,   j  ,      *
             *        j                    *
             */
            j = -2;
    }   }
    return(j);
}

int
pgoblin_cmdexe(int cmd, pgoblin_exenv *exenv, pgoblin_nr *rn) {
#   define blin_internal_flags (exenv->options->flags & BLIN_MASK)
    int ex, i, b;

    ifBLIN_QX3( "+%s %c%c%c%c%c%c %08X"
              , pgoblin.a[cmd].str
              , pgoblin_regn[rn[0]]
              , pgoblin_regn[rn[1]]
              , pgoblin_regn[rn[2]]
              , pgoblin_regn[rn[3]]
              , pgoblin_regn[rn[4]]
              , pgoblin_regn[rn[5]]
              , pgoblin.a[cmd].flags
              );
    for (i = 0; i < pgoblin.symparam; i++) if (0 > (b = pgoblin_breg(cmd, i))) {
        switch(i) {
        case PGO_CSTY:
            break;
        case PGO_CJOB:
            break;
        case PGO_CCON:
            break;
        default:
            break;
    }   }
    ex = (*(pgoblin.a[cmd].car))(exenv, rn);
    ifBLIN_QX3( "- %d", ex);
    return(ex);
#   undef blin_internal_flags
}

void *
pgoblin_getr(pgoblin_main *options, pgoblin_nr rn, int reg) {
#   define blin_internal_flags (options->flags & BLIN_MASK)
    switch(reg) {
    case PGO_CSTY: return(&(options->style[rn]));
    case PGO_CJOB: return(&(options->job[rn]));
    case PGO_CCON: return(&(options->conn[rn]));
    default:       return(&(options->io[rn]));
#   undef blin_internal_flags
}   }

void
pgoblin_go( void *x           /*  (f > 0)   (f < 0)     */
          , int f             /*    pgoblin_syntax.ch[]  *-1 */
          , const char *file  /*                            */
          , const char *func  /*                                   */
          , int line          /*                        */
          ) {
#   define blin_internal_flags (pgoblin_options->flags)
    if  (0 <= pgoblin_options->xtrace) {
        u_int64_t d;
        u_int64_t p;
        u_int64_t q;
        int       i;
        struct {
            u_char    dummy0;  /*                      */
            u_char    dummy1;  /*                      */
            u_char    y0;      /*    pgoblin_syntax.ch[] */
            u_char    y1;      /*                        */
            pid_t     pid;     /* PID                         */
        }         r;

        r.pid = getpid();
        if  (0 < f) {
            switch(f) {
            case PGOBLIN_STYLE   :
            case PGOBLIN_JOB     :
            case PGOBLIN_CONNECT :
            case PGOBLIN_OUTSET  :
            case PGOBLIN_PQRESULT:
            case PGOBLIN_MIFEDES :
            case PGOBLIN_TEXTPARM:
            case PGOBLIN_PATHPARM:
                r.y0 = f;
                r.y1 = (u_int64_t)x;
                break;
            default              :
                r.y0 = 0;
                r.y1 = f;
            }
        } else {
            switch(-f) {
            case PGOBLIN_STYLE   :
                p = (u_int64_t)&pgoblin_options->style[0];
                q = (u_int64_t)&pgoblin_options->style[1];
                break;
            case PGOBLIN_JOB     :
                p = (u_int64_t)&pgoblin_options->job[0];
                q = (u_int64_t)&pgoblin_options->job[1];
                break;
            case PGOBLIN_CONNECT :
                p = (u_int64_t)&pgoblin_options->conn[0];
                q = (u_int64_t)&pgoblin_options->conn[1];
                break;
            case PGOBLIN_OUTSET  :
            case PGOBLIN_PQRESULT:
            case PGOBLIN_MIFEDES :
            case PGOBLIN_TEXTPARM:
            case PGOBLIN_PATHPARM:
                p = (u_int64_t)&pgoblin_options->io[0];
                q = (u_int64_t)&pgoblin_options->io[1];
                break;
            default              :
                r.y0 = 0;
                r.y1 = -f;
                goto out;
            }
            d = ((u_int64_t)x - p) / (q - p);
            if  (((d * (q - p)) != ((u_int64_t)x - p)) || !!(d & ~0x3FUL)) {
                ifBLIN_QX0( "@%s:%s:%d: "
                            "Lost syn %"BLIN_X" i%"BLIN_X" c%"BLIN_X" j%"BLIN_X" s%"BLIN_X" f%02X"
                          , file
                          , func
                          , line
                          , BLIN_I(x)
                          , BLIN_I(&pgoblin_options->io[0])
                          , BLIN_I(&pgoblin_options->conn[0])
                          , BLIN_I(&pgoblin_options->job[0])
                          , BLIN_I(&pgoblin_options->style[0])
                          , -f
                          );
                d = 0x3F;
            }
            r.y0 = -f;
            r.y1 = d;
        }
out:
        errno = 0;
        if  (6 != (i = write(pgoblin_options->xtrace, &r.y0, 2 + sizeof(pid_t)))) {
            ifBLIN_QW0("(%d)write(,,6)=%d", r.pid, i);
    }   }
#   undef blin_internal_flags
}
